"""The cell module representing the NEURON cell and operations on it."""

from .core import Cell
from .cell_dict import CellDict
from .sonata_proxy import SonataProxy
from .ballstick import create_ball_stick



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generates stimuli to be injected into cells."""

import math
import logging

import neuron

from bluecellulab.cell.random import gamma

logger = logging.getLogger(__name__)


def gen_shotnoise_signal(tau_D, tau_R, rate, amp_mean, amp_var,
                         duration, dt=0.25, rng=None):
    """Adds a Poisson shot noise signal with gamma-distributed amplitudes and
    bi-exponential impulse response.

    tau_D: bi-exponential decay time [ms]
    tau_R: bi-exponential rise time [ms]
    rate: Poisson event rate [Hz]
    amp_mean: mean of gamma-distributed amplitudes [nA]
    amp_var: variance of gamma-distributed amplitudes [nA^2]
    duration: duration of signal [ms]
    dt: timestep [ms]
    rng: random number generator object
    """
    if rng is None:
        logger.info("Using a default RNG for shot noise generation")
        rng = neuron.h.Random()  # Creates a default RNG

    tvec = neuron.h.Vector()
    tvec.indgen(0, duration, dt)  # time vector
    ntstep = len(tvec)  # total number of timesteps

    rate_ms = rate / 1000  # rate in 1 / ms [mHz]
    napprox = 1 + int(duration * rate_ms)  # approximate number of events, at least one
    napprox = int(napprox + 3 * math.sqrt(napprox))  # better bound, as in elephant

    exp_scale = 1 / rate  # scale parameter of exponential distribution of time intervals
    rng.negexp(exp_scale)
    iei = neuron.h.Vector(napprox)
    iei.setrand(rng)  # generate inter-event intervals

    ev = neuron.h.Vector()
    ev.integral(iei, 1).mul(1000)  # generate events in ms
    # add events if last event falls short of duration
    while ev[-1] < duration:
        iei_new = neuron.h.Vector(100)  # generate 100 new inter-event intervals
        iei_new.setrand(rng)            # here rng is still negexp
        ev_new = neuron.h.Vector()
        ev_new.integral(iei_new, 1).mul(1000).add(ev[-1])  # generate new shifted events in ms
        ev.append(ev_new)  # append new events
    ev.where("<", duration)  # remove events exceeding duration
    ev.div(dt)  # divide events by timestep

    nev = neuron.h.Vector([round(x) for x in ev])  # round to integer timestep index
    nev.where("<", ntstep)  # remove events exceeding number of timesteps

    sign = 1
    # if amplitude mean is negative, invert sign of current
    if amp_mean < 0:
        amp_mean = -amp_mean
        sign = -1

    gamma_scale = amp_var / amp_mean      # scale parameter of gamma distribution
    gamma_shape = amp_mean / gamma_scale  # shape parameter of gamma distribution
    # sample gamma-distributed amplitudes
    amp = gamma(rng, gamma_shape, gamma_scale, len(nev))

    E = neuron.h.Vector(ntstep, 0)  # full signal
    for n, A in zip(nev, amp):
        E.x[int(n)] += sign * A  # add impulses, may overlap due to rounding to timestep

    # perform equivalent of convolution with bi-exponential impulse response
    # through a composite autoregressive process with impulse train as innovations

    # unitless quantities (time measured in timesteps)
    a = math.exp(-dt / tau_D)
    b = math.exp(-dt / tau_R)
    D = -math.log(a)
    R = -math.log(b)
    t_peak = math.log(R / D) / (R - D)
    A = (a / b - 1) / (a ** t_peak - b ** t_peak)

    P = neuron.h.Vector(ntstep, 0)
    B = neuron.h.Vector(ntstep, 0)

    # composite autoregressive process with exact solution
    # P[n] = b * (a ^ n - b ^ n) / (a - b)
    # for unit response B[0] = P[0] = 0, E[0] = 1
    for n in range(1, ntstep):
        P.x[n] = a * P[n - 1] + b * B[n - 1]
        B.x[n] = b * B[n - 1] + E[n - 1]

    P.mul(A)  # normalize to peak amplitude

    # append zero at end
    tvec.append(duration)
    P.append(.0)

    return tvec, P


def get_relative_shotnoise_params(mean, var, tau_D, tau_R, cv_square):
    """Returns Rate, amp_mean and amp_var parameters."""
    # bi-exponential time to peak [ms]
    t_peak = math.log(tau_D / tau_R) / (1 / tau_R - 1 / tau_D)
    # bi-exponential peak height [1]
    x_peak = math.exp(-t_peak / tau_D) - math.exp(-t_peak / tau_R)

    rate_ms = (1 + cv_square) / 2 * (mean ** 2 / var) / (tau_D + tau_R)
    rate = rate_ms * 1000  # rate in 1 / s [Hz]
    amp_mean = mean * x_peak / rate_ms / (tau_D - tau_R)
    amp_var = cv_square * amp_mean ** 2

    return rate, amp_mean, amp_var


def gen_ornstein_uhlenbeck(tau, sigma, mean, duration, dt=0.25, rng=None):
    """Adds an Ornstein-Uhlenbeck process with given correlation time, standard
    deviation and mean value.

    tau: correlation time [ms], white noise if zero
    sigma: standard deviation [uS]
    mean: mean value [uS]
    duration: duration of signal [ms]
    dt: timestep [ms]
    rng: random number generator object
    """

    if rng is None:
        logger.info("Using a default RNG for Ornstein-Uhlenbeck process")
        rng = neuron.h.Random()  # Creates a default RNG

    tvec = neuron.h.Vector()
    tvec.indgen(0, duration, dt)  # time vector
    ntstep = len(tvec)  # total number of timesteps

    svec = neuron.h.Vector(ntstep, 0)  # stim vector

    noise = neuron.h.Vector(ntstep)  # Gaussian noise
    rng.normal(0.0, 1.0)
    noise.setrand(rng)  # generate Gaussian noise

    if tau < 1e-9:
        svec = noise.mul(sigma)  # white noise
    else:
        mu = math.exp(-dt / tau)  # auxiliar factor [unitless]
        A = sigma * math.sqrt(1 - mu * mu)  # amplitude [uS]
        noise.mul(A)  # scale noise by amplitude [uS]

        # Exact update formula (independent of dt) from Gillespie 1996
        for n in range(1, ntstep):
            svec.x[n] = svec[n - 1] * mu + noise[n]  # signal [uS]

    svec.add(mean)  # shift signal by mean value [uS]

    # append zero at end
    tvec.append(duration)
    svec.append(.0)

    return tvec, svec



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a NEURON Segment in Python (for drawing)."""

import neuron

from bluecellulab.neuron_interpreter import eval_neuron

type_colormap = {'apic': 'm', 'dend': 'r', 'soma': 'k', 'axon': 'b', 'myelin': 'g', 'unknown': 'c'}


class PSegment:
    """A python representation of a segment."""

    def __init__(self, hsegment, parentsection):
        # import matplotlib as plt
        from matplotlib import colormaps

        self.hsegment = hsegment
        self.parentsection = parentsection
        self.L = self.parentsection.L / self.parentsection.hsection.nseg
        self.diam = self.hsegment.diam

        self.figure = None
        self.figX = None
        self.figY = None
        self.color_map = colormaps["hot"]
        self.ax = None
        self.patch = None
        self.plotvariable = None
        self.mechanism_names = [mechanism.name() for mechanism in hsegment]
        self.varbounds = None

    def setupDraw(self, figure, x, y, variable=None, varbounds=None):
        """Set up the drawing of a segment."""
        import matplotlib as plt

        self.figure = figure
        self.plotvariable = variable
        self.varbounds = varbounds
        self.ax = self.figure.gca()
        self.figX = x
        self.figY = y
        self.patch = plt.patches.Rectangle(
            [self.figX, self.figY], self.diam, self.L,
            facecolor=type_colormap[self.parentsection.section_type],
            edgecolor=type_colormap[self.parentsection.section_type])
        self.ax.add_patch(self.patch)

    def redraw(self):
        """Redraw a segment."""
        if self.plotvariable:
            plotvariable_value = self.get_variable_value(self.plotvariable)
            if plotvariable_value is not None:
                self.patch.set_facecolor(self.color_map(
                    (plotvariable_value - self.varbounds[0]) /
                    (self.varbounds[1] - self.varbounds[0])))
            else:
                self.patch.set_facecolor(self.color_map(1.0))
                self.patch.set_hatch("/")
            self.ax.draw_artist(self.patch)

    def get_variable_value(self, variable):
        """Get a variable value in a segment."""
        if variable == "v" or neuron.h.execute1(
            "{%s.%s(%f)}"
            % (
                neuron.h.secname(sec=self.parentsection.hsection),
                variable,
                self.hsegment.x,
            ),
            0,
        ):
            return eval_neuron(f"self.hsegment.{variable}", self=self)
        else:
            return None



"""Controlled simulations in parallel."""

from __future__ import annotations
import multiprocessing
from multiprocessing.pool import Pool
from bluecellulab.simulation.neuron_globals import NeuronGlobals


class IsolatedProcess(Pool):
    """Multiprocessing Pool that restricts a worker to run max 1 process.

    Use this when running isolated NEURON simulations. Running 2 NEURON
    simulations on a single process is to be avoided. Required global
    NEURON simulation parameters will automatically be passed to each
    worker. `fork` mode of multiprocessing is used to pass any other
    global NEURON parameters to each worker.
    """

    def __init__(self, processes: int | None = 1):
        """Initialize the IsolatedProcess pool.

        Args:
            processes: The number of processes to use for running the simulations.
                       If set to None, then the number returned by os.cpu_count() is used.
        """
        ctx = multiprocessing.get_context("fork")
        neuron_global_params = NeuronGlobals.get_instance().export_params()
        super().__init__(
            processes=processes,
            initializer=self.init_worker,
            initargs=(neuron_global_params,),
            maxtasksperchild=1,
            context=ctx,
        )

    @staticmethod
    def init_worker(neuron_global_params):
        """Load global parameters for the NEURON environment in each worker
        process."""
        NeuronGlobals.get_instance().load_params(neuron_global_params)



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that represents a plot window in bluecellulab."""

import itertools


class PlotWindow:
    """Class the represents a plotting window."""

    def __init__(self, var_list, cell, xlim, ylim, title):
        import pylab
        self.cell = cell
        self.var_list = var_list
        pylab.ion()
        self.figure = pylab.figure(figsize=(10, 10))
        # pylab.ioff()

        self.ax = self.figure.gca()
        self.canvas = self.ax.figure.canvas

        self.figure.suptitle(title)
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)
        self.ax.set_xlabel("ms")
        self.ax.set_ylabel("mV")

        self.background = self.canvas.copy_from_bbox(self.ax.bbox)
        self.line = {}

        linenumber = 0
        for var_name in self.var_list:
            recording = self.cell.get_recording(var_name)
            if recording:
                time = self.cell.get_time()
            else:
                time = self.cell.get_time()[1:]

            # print dir(pylab.gca()._get_lines)
            # print pylab.gca()._get_lines.color_cycle
            # Sorry, don't see a way but disable this warning to access this
            colors = pylab.rcParams['axes.prop_cycle'].by_key()['color']

            linecolors = list(itertools.islice(itertools.cycle(colors), 0, 50))

            self.line[var_name] = pylab.Line2D(
                time, recording, label=var_name,
                color=linecolors[linenumber % len(linecolors)])
            self.ax.add_line(self.line[var_name])
            linenumber += 1

        self.ax.legend()

        self.figure.canvas.draw()

        self.drawCount = 10

    def redraw(self):
        """Redraw the plot window."""
        if not self.drawCount:
            time = self.cell.get_time()
            for var_name in self.var_list:
                voltage = self.cell.get_recording(var_name)
                self.line[var_name].set_data(time, voltage)
                self.ax.draw_artist(self.line[var_name])
            self.canvas.blit(self.ax.bbox)
            self.drawCount = 10
        else:
            self.drawCount = self.drawCount - 1



"""Simulation module responsible of NEURON simulation."""

from .simulation import Simulation
from .neuron_globals import set_global_condition_parameters
from .neuron_globals import set_minis_single_vesicle_values



"""Module for injecting a sequence of protocols to the cell."""
from __future__ import annotations
from enum import Enum, auto
from typing import NamedTuple, Sequence, Dict

import neuron
import numpy as np
from bluecellulab.cell.core import Cell
from bluecellulab.cell.template import TemplateParams
from bluecellulab.simulation.parallel import IsolatedProcess
from bluecellulab.simulation.simulation import Simulation
from bluecellulab.stimulus.factory import Stimulus, StimulusFactory


class StimulusName(Enum):
    """Allowed values for the StimulusName."""
    AP_WAVEFORM = auto()
    IDREST = auto()
    IV = auto()
    FIRE_PATTERN = auto()
    POS_CHEOPS = auto()
    NEG_CHEOPS = auto()


class Recording(NamedTuple):
    """A tuple of the current, voltage and time recordings."""
    current: np.ndarray
    voltage: np.ndarray
    time: np.ndarray


StimulusRecordings = Dict[str, Recording]


def run_stimulus(
    template_params: TemplateParams,
    stimulus: Stimulus,
    section: str,
    segment: float,
    cvode: bool = True,
) -> Recording:
    """Creates a cell and stimulates it with a given stimulus.

    Args:
        template_params: The parameters to create the cell from a template.
        stimulus: The input stimulus to inject into the cell.
        section: Name of the section of cell where the stimulus is to be injected.
        segment: The segment of the section where the stimulus is to be injected.
        cvode: True to use variable time-steps. False for fixed time-steps.

    Returns:
        The voltage-time recording at the specified location.

    Raises:
        ValueError: If the time and voltage arrays are not the same length.
    """
    cell = Cell.from_template_parameters(template_params)
    neuron_section = cell.sections[section]
    cell.add_voltage_recording(neuron_section, segment)
    iclamp, _ = cell.inject_current_waveform(
        stimulus.time, stimulus.current, section=neuron_section, segx=segment
    )
    current_vector = neuron.h.Vector()
    current_vector.record(iclamp._ref_i)
    simulation = Simulation(cell)
    simulation.run(stimulus.stimulus_time, cvode=cvode)
    current = np.array(current_vector.to_python())
    voltage = cell.get_voltage_recording(neuron_section, segment)
    time = cell.get_time()
    if len(time) != len(voltage) or len(time) != len(current):
        raise ValueError("Time, current and voltage arrays are not the same length")
    return Recording(current, voltage, time)


def apply_multiple_stimuli(
    cell: Cell,
    stimulus_name: StimulusName,
    amplitudes: Sequence[float],
    threshold_based: bool = True,
    section_name: str | None = None,
    segment: float = 0.5,
    n_processes: int | None = None,
    cvode: bool = True,
) -> StimulusRecordings:
    """Apply multiple stimuli to the cell on isolated processes.

    Args:
        cell: The cell to which the stimuli are applied.
        stimulus_name: The name of the stimulus to apply.
        amplitudes: The amplitudes of the stimuli to apply.
        threshold_based: Whether to consider amplitudes to be
            threshold percentages or to be raw amplitudes.
        section_name: Section name of the cell where the stimuli are applied.
          If None, the stimuli are applied at the soma[0] of the cell.
        segment: The segment of the section where the stimuli are applied.
        n_processes: The number of processes to use for running the stimuli.
        cvode: True to use variable time-steps. False for fixed time-steps.

    Returns:
        A dictionary where the keys are the names of the stimuli and the values
        are the recordings of the cell's response to each stimulus.

    Raises:
        ValueError: If the stimulus name is not recognized.
    """
    res: StimulusRecordings = {}
    stim_factory = StimulusFactory(dt=1.0)
    task_args = []
    section_name = section_name if section_name is not None else "soma[0]"

    # Prepare arguments for each stimulus
    for amplitude in amplitudes:
        if threshold_based:
            thres_perc = amplitude
            amp = None
        else:
            thres_perc = None
            amp = amplitude

        if stimulus_name == StimulusName.AP_WAVEFORM:
            stimulus = stim_factory.ap_waveform(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        elif stimulus_name == StimulusName.IDREST:
            stimulus = stim_factory.idrest(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        elif stimulus_name == StimulusName.IV:
            stimulus = stim_factory.iv(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        elif stimulus_name == StimulusName.FIRE_PATTERN:
            stimulus = stim_factory.fire_pattern(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        elif stimulus_name == StimulusName.POS_CHEOPS:
            stimulus = stim_factory.pos_cheops(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        elif stimulus_name == StimulusName.NEG_CHEOPS:
            stimulus = stim_factory.neg_cheops(
                threshold_current=cell.threshold, threshold_percentage=thres_perc, amplitude=amp
            )
        else:
            raise ValueError("Unknown stimulus name.")

        task_args.append((cell.template_params, stimulus, section_name, segment, cvode))

    with IsolatedProcess(processes=n_processes) as pool:
        # Map expects a function and a list of argument tuples
        results = pool.starmap(run_stimulus, task_args)

    # Associate each result with a key
    for amplitude, result in zip(amplitudes, results):
        key = f"{stimulus_name}_{amplitude}"
        res[key] = result

    return res
